@@ -12,7 +12,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from typing import Iterable, Any, Tuple, Dict, BinaryIO, Union, List, cast
+from typing import Iterable, Any, Tuple, Dict, Callable, BinaryIO, Union, List, cast
 from io import BytesIO, BufferedReader
 from collections.abc import Mapping, Sequence
 import msgpack
@@ -22,8 +22,27 @@ import json
 import math
 import re
 
-from nomad import utils
+from nomad import utils, config, infrastructure
+from nomad.metainfo import Quantity, Reference, Definition, MSection, Section, SubSection
+from nomad.datamodel import EntryArchive
+from nomad.datamodel.metainfo.public import fast_access
 
+'''
+The archive storage is made from two tiers. First the whole archive is stored in
+files, secondly parts of the archive are stored in mongodb documents.
+
+The file storage is done in msg-pack files. Each file contains the archive of many
+entries (all entries of an upload). These msg-pack files contain the JSON serialized
+version of the metainfo archive (see module:`nomad.metainfo`). In addition msg-pack
+contains TOC information for quicker access of individual sections. See :func:`write_archive`
+and :func:`read_archvive`. In addition there query functionality to partially read
+specified sections from an archive: func:`query_archive`.
+
+The mongo storage uses mongodb's native bson to store JSON serialized metainfo archive
+data. Each document in mongodb holds the partial archive of single entry. Which parts
+of an archive are stored in mongo is determined by the metainfo and
+section annotations/categories.
+'''
 
 __packer = msgpack.Packer(autoreset=True, use_bin_type=True)
 
@@ -389,7 +408,7 @@ class ArchiveReader(ArchiveObject):
                 raise KeyError(key)
 
             positions = self._toc.get(key)
-            # TODO use hash algo instead of binary search
+            # TODO use hash algorithm instead of binary search
             if positions is None:
                 r_start = 0
                 r_end = self._n_toc
@@ -426,14 +445,14 @@ class ArchiveReader(ArchiveObject):
 
     def __iter__(self):
         if self.toc_entry is None:
-            # is not necessarely read when using blocked toc
+            # is not necessarily read when using blocked toc
             self.toc_entry = self._read(self.toc_position)
 
         return self.toc_entry.__iter__()
 
     def __len__(self):
         if self.toc_entry is None:
-            # is not necessarely read when using blocked toc
+            # is not necessarily read when using blocked toc
             self.toc_entry = self._read(self.toc_position)
 
         return self.toc_entry.__len__()
@@ -499,7 +518,7 @@ def write_archive(
     The top-level TOC positions are 2*5byte encoded integers. This will give the top-level TOC a
     predictable layout and will allow to partially read this TOC.
 
-    The TOC of each entry will have the same structure than the data upto a certain
+    The TOC of each entry will have the same structure than the data up to a certain
     TOC depth. A TOC object will hold the position of the object it refers to (key 'pos')
     and further deeper TOC data (key 'toc'). Only data objects (dict instances) will
     have TOC objects and only object count towards the TOC depth. Positions in the entry
@@ -528,7 +547,7 @@ def read_archive(file_or_path: str, **kwargs) -> ArchiveReader:
 
     Returns:
         A mapping (dict-like) that can be used to access the archive data. The mapping
-        will lazyly load data as it is used. The mapping needs to be closed or used within
+        will lazily load data as it is used. The mapping needs to be closed or used within
         a 'with' statement to free the underlying file resource after use.
     '''
 
@@ -538,15 +557,14 @@ def read_archive(file_or_path: str, **kwargs) -> ArchiveReader:
 __query_archive_key_pattern = re.compile(r'^([\s\w\-]+)(\[([-?0-9]*)(:([-?0-9]*))?\])?$')
 
 
-def query_archive(f_or_archive_reader: Union[str, ArchiveReader, BytesIO], query_dict: dict, **kwargs):
-    def _fix_index(index, length):
-        if index is None:
-            return index
-        if index < 0:
-            return max(-(length), index)
-        else:
-            return min(length, index)
-
+def query_archive(
+        f_or_archive_reader: Union[str, ArchiveReader, BytesIO], query_dict: dict,
+        **kwargs) -> Dict:
+    '''
+    Takes an open msg-pack based archive (either as str, reader, or BytesIO) and returns
+    the archive as JSON serializable dictionary filtered based on the given required
+    specification.
+    '''
     def _to_son(data):
         if isinstance(data, (ArchiveList, List)):
             data = [_to_son(item) for item in data]
@@ -556,89 +574,341 @@ def query_archive(f_or_archive_reader: Union[str, ArchiveReader, BytesIO], query
 
         return data
 
-    def _load_data(query_dict: Dict[str, Any], archive_item: ArchiveObject, main_section: bool = False):
-        if not isinstance(query_dict, dict):
-            return _to_son(archive_item)
-
-        result = {}
-        for key, val in query_dict.items():
-            key = key.strip()
-
-            # process array indices
-            match = __query_archive_key_pattern.match(key)
-            index: Union[Tuple[int, int], int] = None
-            if match:
-                key = match.group(1)
-
-                # check if we have indices
-                if match.group(2) is not None:
-                    first_index, last_index = None, None
-                    group = match.group(3)
-                    first_index = None if group == '' else int(group)
-
-                    if match.group(4) is not None:
-                        group = match.group(5)
-                        last_index = None if group == '' else int(group)
-                        index = (0 if first_index is None else first_index, last_index)
-
-                    else:
-                        index = first_index  # one item
-
-                else:
-                    index = None
-
-            else:
-                raise ArchiveQueryError('invalid key format: %s' % key)
-
-            # support for shorter uuids
-            if main_section:
-                archive_key = adjust_uuid_size(key)
-            else:
-                archive_key = key
-
-            try:
-                archive_child = archive_item[archive_key]
-                is_list = isinstance(archive_child, (ArchiveList, list))
-
-                if index is None and is_list:
-                    index = (0, None)
-                elif index is not None and not is_list:
-                    raise ArchiveQueryError('cannot use list key on none list %s' % key)
-
-                if index is None:
-                    pass
-                else:
-                    length = len(archive_child)
-                    if isinstance(index, tuple):
-                        index = (_fix_index(index[0], length), _fix_index(index[1], length))
-                        if index[0] == index[1]:
-                            archive_child = [archive_child[index[0]]]
-                        else:
-                            archive_child = archive_child[index[0]: index[1]]
-                    else:
-                        archive_child = [archive_child[_fix_index(index, length)]]
-
-                if isinstance(archive_child, (ArchiveList, list)):
-                    result[key] = [_load_data(val, item) for item in archive_child]
-                else:
-                    result[key] = _load_data(val, archive_child)
-
-            except (KeyError, IndexError):
-                continue
-
-        return result
+    def _load_data(query_dict: Dict[str, Any], archive_item: ArchiveObject) -> Dict:
+        query_dict_with_fixed_ids = {
+            adjust_uuid_size(key): value for key, value in query_dict.items()}
+        return filter_archive(query_dict_with_fixed_ids, archive_item, transform=_to_son)
 
     if isinstance(f_or_archive_reader, ArchiveReader):
-        return _load_data(query_dict, f_or_archive_reader, True)
+        return _load_data(query_dict, f_or_archive_reader)
 
     elif isinstance(f_or_archive_reader, (BytesIO, str)):
         with ArchiveReader(f_or_archive_reader, **kwargs) as archive:
-            return _load_data(query_dict, archive, True)
+            return _load_data(query_dict, archive)
 
     else:
         raise TypeError('%s is neither a file-like nor ArchiveReader' % f_or_archive_reader)
 
 
+def filter_archive(
+        required: Dict[str, Any], archive_item: Union[Dict, ArchiveObject],
+        transform: Callable) -> Dict:
+
+    def _fix_index(index, length):
+        if index is None:
+            return index
+        if index < 0:
+            return max(-(length), index)
+        else:
+            return min(length, index)
+
+    if archive_item is None:
+        return None
+
+    if not isinstance(required, dict):
+        return transform(archive_item)
+
+    result: Dict[str, Any] = {}
+    for key, val in required.items():
+        key = key.strip()
+
+        # process array indices
+        match = __query_archive_key_pattern.match(key)
+        index: Union[Tuple[int, int], int] = None
+        if match:
+            key = match.group(1)
+
+            # check if we have indices
+            if match.group(2) is not None:
+                first_index, last_index = None, None
+                group = match.group(3)
+                first_index = None if group == '' else int(group)
+
+                if match.group(4) is not None:
+                    group = match.group(5)
+                    last_index = None if group == '' else int(group)
+                    index = (0 if first_index is None else first_index, last_index)
+
+                else:
+                    index = first_index  # one item
+
+            else:
+                index = None
+
+        else:
+            raise ArchiveQueryError('invalid key format: %s' % key)
+
+        try:
+            archive_child = archive_item[key]
+            is_list = isinstance(archive_child, (ArchiveList, list))
+
+            if index is None and is_list:
+                index = (0, None)
+
+            elif index is not None and not is_list:
+                raise ArchiveQueryError('cannot use list key on none list %s' % key)
+
+            if index is None:
+                pass
+            else:
+                length = len(archive_child)
+                if isinstance(index, tuple):
+                    index = (_fix_index(index[0], length), _fix_index(index[1], length))
+                    if index[0] == index[1]:
+                        archive_child = [archive_child[index[0]]]
+                    else:
+                        archive_child = archive_child[index[0]: index[1]]
+                else:
+                    archive_child = [archive_child[_fix_index(index, length)]]
+
+            if isinstance(archive_child, (ArchiveList, list)):
+                result[key] = [
+                    filter_archive(val, item, transform=transform)
+                    for item in archive_child]
+            else:
+                result[key] = filter_archive(val, archive_child, transform=transform)
+
+        except (KeyError, IndexError):
+            continue
+
+    return result
+
+
+def create_partial_archive(archive: EntryArchive) -> Dict:
+    '''
+    Creates a partial archive JSON serializable dict that can be stored directly.
+    The given archive is filtered based on the metainfo category ``fast_access``.
+    Selected sections and other data that they reference (recursively) comprise the
+    resulting partial archive.
+
+    TODO at the moment is hard coded and NOT informed by the metainfo. We simply
+    add sections EntryMetadata and Workflow.
+
+    Arguments:
+        archive: The archive as an :class:`EntryArchive` instance.
+
+    Returns: the partial archive in JSON serializable dictionary form.
+    '''
+    # A list with all referenced sections that might not yet been ensured to be in the
+    # resulting partial archive
+    referenceds: List[MSection] = []
+    # contents keeps track of all sections in the partial archive by keeping their
+    # JSON serializable form and placeholder status in a dict
+    contents: Dict[MSection, Tuple[dict, bool]] = dict()
+
+    def partial(definition: Definition, section: MSection) -> bool:
+        '''
+        ``m_to_dict`` partial function that selects what goes into the partial archive
+        and what not. It also collects references as a side-effect.
+        '''
+        if section.m_def == EntryArchive.m_def:
+            if definition.m_def == Quantity:
+                return True
+            return fast_access.m_def in definition.categories
+
+        if isinstance(definition, Quantity) and isinstance(definition.type, Reference) \
+                and fast_access.m_def in definition.categories:
+            # Reference list in partial archives are not supported
+            if definition.is_scalar:
+                referenced = getattr(section, definition.name)
+                if referenced is not None and referenced not in contents:
+                    referenceds.append(referenced)
+
+        if isinstance(definition, SubSection):
+            return fast_access.m_def in definition.categories
+
+        return True
+
+    # add the main content
+    partial_contents = archive.m_to_dict(partial=partial)
+
+    # add the referenced data
+    def add(section, placeholder=False) -> dict:
+        '''
+        Adds the given section to partial_contents at the right place. If not a placeholder,
+        the section's serialization is added (or replacing an existing placeholder).
+        Otherwise, an empty dict is added as a placeholder for referenced children.
+        '''
+        result: Dict[str, Any] = None
+        content, content_is_placeholder = contents.get(section, (None, True))
+        if content is not None:
+            if content_is_placeholder and not placeholder:
+                # the placeholder gets replaced later
+                pass
+            else:
+                return content
+
+        if section.m_parent is None:
+            contents[section] = partial_contents, False
+            return partial_contents
+
+        parent_dict = add(section.m_parent, placeholder=True)
+        if placeholder:
+            result = {}
+        else:
+            result = section.m_to_dict(partial=partial)
+
+        sub_section = section.m_parent_sub_section
+        if sub_section.repeats:
+            sections = parent_dict.setdefault(sub_section.name, [])
+            while len(sections) < section.m_parent_index + 1:
+                sections.append(None)
+            sections[section.m_parent_index] = result
+        else:
+            parent_dict[sub_section.name] = result
+
+        contents[section] = result, placeholder
+        return result
+
+    # we add referenced objects as long as they are added by subsequent serialization
+    # of referenced sections to implement the recursive nature of further references in
+    # already referenced sections.
+    while len(referenceds) > 0:
+        referenced = referenceds.pop()
+        add(referenced)
+
+    return partial_contents
+
+
+def write_partial_archive_to_mongo(archive: EntryArchive):
+    ''' Partially writes the given archive to mongodb. '''
+    mongo_db = infrastructure.mongo_client[config.mongo.db_name]
+    mongo_collection = mongo_db['archive']
+    mongo_id = archive.section_metadata.calc_id
+
+    partial_archive_dict = create_partial_archive(archive)
+    partial_archive_dict['_id'] = mongo_id
+    mongo_collection.replace_one(dict(_id=mongo_id), partial_archive_dict, upsert=True)
+
+
+def read_partial_archive_from_mongo(entry_id: str, as_dict=False) -> Union[EntryArchive, Dict]:
+    '''
+    Reads the partial archive for the given id from mongodb.
+
+    Arguments:
+        entry_id: The entry id for the entry.
+        as_dict: Return the JSON serializable dictionary form of the archive not the
+            :class:`EntryArchive` form.
+    '''
+    mongo_db = infrastructure.mongo_client[config.mongo.db_name]
+    mongo_collection = mongo_db['archive']
+    archive_dict = mongo_collection.find_one(dict(_id=entry_id))
+
+    if as_dict:
+        return archive_dict
+
+    return EntryArchive.m_from_dict(archive_dict)
+
+
+def read_partial_archives_from_mongo(entry_ids: List[str], as_dict=False) -> Dict[str, Union[EntryArchive, Dict]]:
+    '''
+    Reads the partial archives for a set of entries of the same upload.
+
+    Arguments:
+        entry_ids: A list of entry ids.
+        as_dict: Return the JSON serializable dictionary form of the archive not the
+            :class:`EntryArchive` form.
+
+    Returns:
+        A dictionary with entry_ids as keys.
+    '''
+    mongo_db = infrastructure.mongo_client[config.mongo.db_name]
+    mongo_collection = mongo_db['archive']
+    archive_dicts = mongo_collection.find(dict(_id={'$in': entry_ids}))
+
+    if as_dict:
+        return {archive_dict.pop('_id'): archive_dict for archive_dict in archive_dicts}
+
+    return {
+        archive_dict.pop('_id'): EntryArchive.m_from_dict(archive_dict)
+        for archive_dict in archive_dicts}
+
+
+__all_parent_sections: Dict[Section, Tuple[str, Section]] = {}
+
+
+def _all_parent_sections():
+    if len(__all_parent_sections) == 0:
+        def add(section):
+            for sub_section in section.all_sub_sections.values():
+                sub_section_section = sub_section.sub_section.m_resolved()
+                __all_parent_sections.setdefault(sub_section_section, []).append((sub_section.qualified_name(), section, ))
+                add(sub_section_section)
+
+        add(EntryArchive.m_def)
+
+    return __all_parent_sections
+
+
+class _Incomplete(Exception): pass
+
+
+def compute_required_with_referenced(required):
+    '''
+    Updates the given required dictionary to ensure that references to non required
+    sections within a partial fast access archive are included. Only references that
+    are directly contained in required are added. References from wildcarded sub sections
+    are ignored.
+
+    Returns: A new required dict or None. None is returned if it is unclear if the required
+    is only accessing information of fast access partial archives.
+    '''
+    # TODO this function should be based on the metainfo
+
+    if not isinstance(required, dict):
+        return required
+
+    if any(key.startswith('section_run') for key in required):
+        return None
+
+    required = dict(**required)
+
+    def add_parent_section(section, child_required):
+        parent_sections = _all_parent_sections().get(section, [])
+        if len(parent_sections) == 0:
+            return [required]
+
+        result = []
+        for name, parent_section in parent_sections:
+            child_key = name.split('.')[-1]
+            for parent_required in add_parent_section(parent_section, None):
+                result.append(parent_required.setdefault(child_key, child_required if child_required else {}))
+
+        return result
+
+    def traverse(
+            current: Union[dict, str],
+            parent: Section = EntryArchive.m_def):
+
+        if isinstance(current, str):
+            return
+
+        for key, value in current.items():
+            prop = key.split('[')[0]
+            prop_definition = parent.all_properties[prop]
+            if isinstance(prop_definition, SubSection):
+                if fast_access.m_def not in prop_definition.categories:
+                    raise _Incomplete()
+
+                traverse(value, prop_definition.sub_section)
+            if isinstance(prop_definition, Quantity) and isinstance(prop_definition.type, Reference):
+                current[prop] = '*'
+                if fast_access.m_def not in prop_definition.categories:
+                    continue
+
+                target_section_def = prop_definition.type.target_section_def.m_resolved()
+                add_parent_section(target_section_def, value)
+                traverse(value, target_section_def)
+
+    try:
+        traverse(dict(**required))
+    except _Incomplete:
+        # We realized that something is required that is not in the partial archive
+        return None
+
+    return required
+
+
 if __name__ == '__main__':
 
     def benchmark():
