@@ -280,12 +280,15 @@ from .metainfo import (
     Environment,
     MEnum,
     Datetime,
+    Capitalized,
     MProxy,
     MetainfoError,
     DeriveError,
     MetainfoReferenceError,
     DataType,
     Reference,
+    Datetime,
+    JSON,
     MResource,
     m_package,
     Annotation,
